from functools import lru_cache

from arcade_core.config_model import Config


@lru_cache(maxsize=1)
def get_config() -> Config:
    """
    Get the Arcade configuration.

    This function is cached, so subsequent calls will return the same Config object
    without reloading from the file, unless the cache is cleared.

    remember to clear the cache if the configuration file is modified.
    use `get_config.cache_clear()` to clear the cache.

    Returns:
        Config: The Arcade configuration.
    """
    return Config.load_from_file()


config = get_config()



"""Tests for MCP Settings."""

import pytest
from pydantic import ValidationError

from arcade_mcp_server.settings import MCPSettings, ServerSettings


class TestServerSettings:
    """Test ServerSettings class."""

    def test_server_settings_defaults(self):
        """Test ServerSettings default values."""
        settings = ServerSettings()

        assert settings.name == "ArcadeMCP"
        assert settings.version == "0.1.0"
        assert settings.title == "ArcadeMCP"
        assert settings.instructions is not None
        assert "available tools" in settings.instructions.lower()

    def test_server_settings_custom_values(self):
        """Test ServerSettings with custom values."""
        settings = ServerSettings(
            name="CustomServer",
            version="2.0.0",
            title="Custom Title",
            instructions="Custom instructions",
        )

        assert settings.name == "CustomServer"
        assert settings.version == "2.0.0"
        assert settings.title == "Custom Title"
        assert settings.instructions == "Custom instructions"

    def test_server_settings_partial_values(self):
        """Test ServerSettings with partial custom values."""
        settings = ServerSettings(
            name="PartialServer",
            version="1.5.0",
        )

        assert settings.name == "PartialServer"
        assert settings.version == "1.5.0"
        assert settings.title == "ArcadeMCP"  # Default value
        assert settings.instructions is not None  # Default value


class TestMCPSettings:
    """Test MCPSettings class."""

    def test_mcp_settings_defaults(self):
        """Test MCPSettings default values."""
        settings = MCPSettings()

        assert settings.server.name == "ArcadeMCP"
        assert settings.server.version == "0.1.0"
        assert settings.server.title == "ArcadeMCP"
        assert settings.server.instructions is not None

    def test_mcp_settings_with_custom_server(self):
        """Test MCPSettings with custom ServerSettings."""
        server_settings = ServerSettings(
            name="TestServer",
            version="3.0.0",
            title="Test Title",
            instructions="Test instructions",
        )
        settings = MCPSettings(server=server_settings)

        assert settings.server.name == "TestServer"
        assert settings.server.version == "3.0.0"
        assert settings.server.title == "Test Title"
        assert settings.server.instructions == "Test instructions"

    def test_mcp_settings_from_env(self, monkeypatch):
        """Test MCPSettings.from_env() uses environment variables."""
        monkeypatch.setenv("MCP_SERVER_NAME", "EnvServer")
        monkeypatch.setenv("MCP_SERVER_VERSION", "4.0.0")
        monkeypatch.setenv("MCP_SERVER_TITLE", "Env Title")
        monkeypatch.setenv("MCP_SERVER_INSTRUCTIONS", "Env instructions")

        settings = MCPSettings.from_env()

        assert settings.server.name == "EnvServer"
        assert settings.server.version == "4.0.0"
        assert settings.server.title == "Env Title"
        assert settings.server.instructions == "Env instructions"


class TestServerSettingsTitleDefault:
    """Test that the default title value is 'ArcadeMCP'."""

    def test_title_default_value(self):
        """Test that the default title value is 'ArcadeMCP'."""
        settings = ServerSettings()
        assert settings.title == "ArcadeMCP"

    def test_title_field_default(self):
        """Test that the title field default is 'ArcadeMCP'."""
        field_info = ServerSettings.model_fields["title"]
        assert field_info.default == "ArcadeMCP"


class TestServerSettingsVersionValidation:
    """Tests for ServerSettings version validation (semver enforcement)."""

    def test_server_settings_rejects_invalid_version(self) -> None:
        """Test ServerSettings raises ValidationError for invalid version."""
        with pytest.raises(ValidationError, match="semver"):
            ServerSettings(version="bad")

    def test_server_settings_accepts_valid_semver(self) -> None:
        """Test ServerSettings accepts valid semver."""
        settings = ServerSettings(version="1.2.3-alpha.1+build.456")
        assert settings.version == "1.2.3-alpha.1+build.456"

    def test_server_settings_normalizes_short_version(self) -> None:
        """Test ServerSettings normalizes MAJOR.MINOR to MAJOR.MINOR.0."""
        settings = ServerSettings(version="1.0")
        assert settings.version == "1.0.0"

    def test_server_settings_normalizes_v_prefix(self) -> None:
        """Test ServerSettings strips v prefix and normalizes the version."""
        settings = ServerSettings(version="v1.0.0")
        assert settings.version == "1.0.0"

    def test_server_settings_normalizes_v_prefix_short(self) -> None:
        """Test ServerSettings strips v prefix from short versions."""
        settings = ServerSettings(version="v1.0")
        assert settings.version == "1.0.0"

    def test_server_settings_normalizes_major_only(self) -> None:
        """Test ServerSettings normalizes MAJOR to MAJOR.0.0."""
        settings = ServerSettings(version="1")
        assert settings.version == "1.0.0"

    def test_server_settings_normalizes_v_major_only(self) -> None:
        """Test ServerSettings strips v prefix from major-only versions."""
        settings = ServerSettings(version="v1")
        assert settings.version == "1.0.0"

    def test_mcp_settings_env_rejects_invalid_version(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test MCP_SERVER_VERSION env var is validated."""
        monkeypatch.setenv("MCP_SERVER_VERSION", "not-valid")
        with pytest.raises(ValidationError, match="semver"):
            MCPSettings.from_env()



from arcade_mcp_server import Context, tool


@tool
async def reporting_progress(context: Context) -> str:
    """Report progress back to the client"""
    total = 5

    for i in range(total):
        await context.progress.report(i + 1, total=total, message=f"Step {i + 1} of {total}")

    return "All progress reported successfully"



import traceback
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any


class ErrorKind(str, Enum):
    """Error kind that is comprised of
    - the who (toolkit, tool, upstream)
    - the when (load time, definition parsing time, runtime)
    - the what (bad_definition, bad_input, bad_output, retry, context_required, fatal, etc.)"""

    TOOLKIT_LOAD_FAILED = "TOOLKIT_LOAD_FAILED"
    TOOL_DEFINITION_BAD_DEFINITION = "TOOL_DEFINITION_BAD_DEFINITION"
    TOOL_DEFINITION_BAD_INPUT_SCHEMA = "TOOL_DEFINITION_BAD_INPUT_SCHEMA"
    TOOL_DEFINITION_BAD_OUTPUT_SCHEMA = "TOOL_DEFINITION_BAD_OUTPUT_SCHEMA"
    TOOL_RUNTIME_BAD_INPUT_VALUE = "TOOL_RUNTIME_BAD_INPUT_VALUE"
    TOOL_RUNTIME_BAD_OUTPUT_VALUE = "TOOL_RUNTIME_BAD_OUTPUT_VALUE"
    TOOL_RUNTIME_RETRY = "TOOL_RUNTIME_RETRY"
    TOOL_RUNTIME_CONTEXT_REQUIRED = "TOOL_RUNTIME_CONTEXT_REQUIRED"
    TOOL_RUNTIME_FATAL = "TOOL_RUNTIME_FATAL"
    UPSTREAM_RUNTIME_BAD_REQUEST = "UPSTREAM_RUNTIME_BAD_REQUEST"
    UPSTREAM_RUNTIME_AUTH_ERROR = "UPSTREAM_RUNTIME_AUTH_ERROR"
    UPSTREAM_RUNTIME_NOT_FOUND = "UPSTREAM_RUNTIME_NOT_FOUND"
    UPSTREAM_RUNTIME_VALIDATION_ERROR = "UPSTREAM_RUNTIME_VALIDATION_ERROR"
    UPSTREAM_RUNTIME_RATE_LIMIT = "UPSTREAM_RUNTIME_RATE_LIMIT"
    UPSTREAM_RUNTIME_SERVER_ERROR = "UPSTREAM_RUNTIME_SERVER_ERROR"
    UPSTREAM_RUNTIME_UNMAPPED = "UPSTREAM_RUNTIME_UNMAPPED"
    UNKNOWN = "UNKNOWN"


class ToolkitError(Exception, ABC):
    """
    Base class for all Arcade errors.

    Note: This class is an abstract class and cannot be instantiated directly.

    These errors are ultimately converted to the ToolCallError schema.
    Attributes expected from subclasses:
      message                   : str                    # user-facing error message
      kind                      : ErrorKind              # the error kind
      can_retry                 : bool                   # whether the operation can be retried
      developer_message         : str | None             # developer-facing error details
      status_code               : int | None             # HTTP status code when relevant
      additional_prompt_content : str | None             # content for retry prompts
      retry_after_ms            : int | None             # milliseconds to wait before retry
      stacktrace                : str | None             # stacktrace information
      extra                     : dict[str, Any] | None  # arbitrary structured metadata
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> "ToolkitError":
        abs_methods = getattr(cls, "__abstractmethods__", None)
        if abs_methods:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__}")
        return super().__new__(cls)

    @abstractmethod
    def create_message_prefix(self, name: str) -> str:
        pass

    def with_context(self, name: str) -> "ToolkitError":
        """
        Add context to the error message.

        Args:
            name: The name of the tool or toolkit that caused the error.

        Returns:
            The error with the context added to the message.
        """
        prefix = self.create_message_prefix(name)
        self.message = f"{prefix}{self.message}"  # type: ignore[has-type]
        if hasattr(self, "developer_message") and self.developer_message:  # type: ignore[has-type]
            self.developer_message = f"{prefix}{self.developer_message}"  # type: ignore[has-type]

        return self

    @property
    def is_toolkit_error(self) -> bool:
        """Check if this error originated from loading a toolkit."""
        return hasattr(self, "kind") and self.kind.name.startswith("TOOLKIT_")

    @property
    def is_tool_error(self) -> bool:
        """Check if this error originated from a tool."""
        return hasattr(self, "kind") and self.kind.name.startswith("TOOL_")

    @property
    def is_upstream_error(self) -> bool:
        """Check if this error originated from an upstream service."""
        return hasattr(self, "kind") and self.kind.name.startswith("UPSTREAM_")

    def __str__(self) -> str:
        return self.message


class ToolkitLoadError(ToolkitError):
    """
    Raised while importing / loading a toolkit package
    (e.g. missing dependency, SyntaxError in module top-level code).
    """

    kind: ErrorKind = ErrorKind.TOOLKIT_LOAD_FAILED
    can_retry: bool = False

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def create_message_prefix(self, toolkit_name: str) -> str:
        return f"[{self.kind.value}] {type(self).__name__} when loading toolkit '{toolkit_name}': "


class ToolError(ToolkitError):
    """
    Any error related to an Arcade tool.

    Note: This class is an abstract class and cannot be instantiated directly.
    """


# ------  definition-time errors (tool developer's responsibility) ------
class ToolDefinitionError(ToolError):
    """
    Raised when there is an error in the definition/signature of a tool.
    """

    kind: ErrorKind = ErrorKind.TOOL_DEFINITION_BAD_DEFINITION

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def create_message_prefix(self, tool_name: str) -> str:
        return f"[{self.kind.value}] {type(self).__name__} in definition of tool '{tool_name}': "


class ToolInputSchemaError(ToolDefinitionError):
    """Raised when there is an error in the schema of a tool's input parameter."""

    kind: ErrorKind = ErrorKind.TOOL_DEFINITION_BAD_INPUT_SCHEMA


class ToolOutputSchemaError(ToolDefinitionError):
    """Raised when there is an error in the schema of a tool's output parameter."""

    kind: ErrorKind = ErrorKind.TOOL_DEFINITION_BAD_OUTPUT_SCHEMA


# ------  runtime errors ------
class ToolRuntimeError(ToolError, RuntimeError):
    """
    Any failure starting from when the tool call begins until the tool call returns.

    Note: This class should typically not be instantiated directly, but rather subclassed.
    """

    kind: ErrorKind = ErrorKind.TOOL_RUNTIME_FATAL
    can_retry: bool = False
    status_code: int | None = None
    extra: dict[str, Any] | None = None

    def __init__(
        self,
        message: str,
        developer_message: str | None = None,
        *,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message)
        self.message = message
        self.developer_message = developer_message  # type: ignore[assignment]
        self.extra = extra

    def create_message_prefix(self, tool_name: str) -> str:
        return f"[{self.kind.value}] {type(self).__name__} during execution of tool '{tool_name}': "

    def stacktrace(self) -> str | None:
        if self.__cause__:
            return "\n".join(traceback.format_exception(self.__cause__))
        return None

    def traceback_info(self) -> str | None:
        """DEPRECATED: Use stacktrace() instead.

        This method is deprecated and will be removed in a future major version.
        """
        return self.stacktrace()

    # wire-format helper
    def to_payload(self) -> dict[str, Any]:
        return {
            "message": self.message,
            "developer_message": self.developer_message,
            "kind": self.kind,
            "can_retry": self.can_retry,
            "status_code": self.status_code,
            **(self.extra or {}),
        }


# 1. ------  serialization errors ------
class ToolSerializationError(ToolRuntimeError):
    """
    Raised when there is an error serializing/marshalling the tool call arguments or return value.

    Note: This class is not intended to be instantiated directly, but rather subclassed.
    """


class ToolInputError(ToolSerializationError):
    """
    Raised when there is an error parsing a tool call argument.
    """

    kind: ErrorKind = ErrorKind.TOOL_RUNTIME_BAD_INPUT_VALUE
    status_code: int = 400


class ToolOutputError(ToolSerializationError):
    """
    Raised when there is an error serializing a tool call return value.
    """

    kind: ErrorKind = ErrorKind.TOOL_RUNTIME_BAD_OUTPUT_VALUE
    status_code: int = 500


# 2. ------  tool-body errors ------
class ToolExecutionError(ToolRuntimeError):
    """
    Raised when an error occurs during the execution of a tool's body.

    This is the base class for tool execution errors and should be raised when catching
    exceptions within a tool to provide meaningful information to the tool caller.
    """

    def __init__(
        self,
        message: str,
        developer_message: str | None = None,
        *,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, developer_message=developer_message, extra=extra)


class RetryableToolError(ToolExecutionError):
    """
    Raised when a tool execution error is retryable.
    """

    kind: ErrorKind = ErrorKind.TOOL_RUNTIME_RETRY
    can_retry: bool = True

    def __init__(
        self,
        message: str,
        developer_message: str | None = None,
        additional_prompt_content: str | None = None,  # TODO: Make required in next major version
        retry_after_ms: int | None = None,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, developer_message=developer_message, extra=extra)
        self.additional_prompt_content = additional_prompt_content
        self.retry_after_ms = retry_after_ms


class ContextRequiredToolError(ToolExecutionError):
    """
    Raised when the combination of additional content from the tool AND
    additional context from the end-user/orchestrator is required before retrying the tool.

    This is typically used when an argument provided to the tool is invalid in some way,
    and immediately prompting an LLM to retry the tool call is not desired.
    """

    kind: ErrorKind = ErrorKind.TOOL_RUNTIME_CONTEXT_REQUIRED

    def __init__(
        self,
        message: str,
        additional_prompt_content: str,
        developer_message: str | None = None,
        *,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, developer_message=developer_message, extra=extra)
        self.additional_prompt_content = additional_prompt_content


class FatalToolError(ToolExecutionError):
    """
    Raised when there is an unexpected or unknown error executing a tool.
    """

    status_code: int = 500

    def __init__(
        self,
        message: str,
        developer_message: str | None = None,
        *,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, developer_message=developer_message, extra=extra)


# 3. ------  upstream errors in tool body------
class UpstreamError(ToolExecutionError):
    """
    Error from an upstream service/API during tool execution.

    This class handles all upstream failures except rate limiting.
    The status_code and extra dict provide details about the specific error type.
    """

    def __init__(
        self,
        message: str,
        developer_message: str | None = None,
        *,
        status_code: int,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, developer_message=developer_message, extra=extra)
        self.status_code = status_code
        # Determine retryability based on status code
        self.can_retry = status_code >= 500 or status_code == 429
        # Set appropriate error kind based on status
        if status_code in (401, 403):
            self.kind = ErrorKind.UPSTREAM_RUNTIME_AUTH_ERROR
        elif status_code == 404:
            self.kind = ErrorKind.UPSTREAM_RUNTIME_NOT_FOUND
        elif status_code == 429:
            self.kind = ErrorKind.UPSTREAM_RUNTIME_RATE_LIMIT
        elif status_code >= 500:
            self.kind = ErrorKind.UPSTREAM_RUNTIME_SERVER_ERROR
        elif 400 <= status_code < 500:
            self.kind = ErrorKind.UPSTREAM_RUNTIME_BAD_REQUEST
        else:
            self.kind = ErrorKind.UPSTREAM_RUNTIME_UNMAPPED


class UpstreamRateLimitError(UpstreamError):
    """
    Rate limit error from an upstream service.

    Special case of UpstreamError that includes retry_after_ms information.
    """

    kind: ErrorKind = ErrorKind.UPSTREAM_RUNTIME_RATE_LIMIT
    can_retry: bool = True

    def __init__(
        self,
        message: str,
        retry_after_ms: int,
        developer_message: str | None = None,
        *,
        extra: dict[str, Any] | None = None,
    ):
        super().__init__(message, status_code=429, developer_message=developer_message, extra=extra)
        self.retry_after_ms = retry_after_ms



"""
Capture mode for EvalSuite.

Capture mode runs evaluation cases and records tool calls from the model
without scoring or evaluating them. This is useful for:
- Generating expected tool calls for new test cases
- Debugging model behavior
- Creating baseline recordings
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from openai import AsyncOpenAI

if TYPE_CHECKING:
    from arcade_evals.eval import EvalSuite


@dataclass
class CapturedToolCall:
    """
    A captured tool call from the model during capture mode.

    Attributes:
        name: The name of the tool that was called.
        args: The arguments passed to the tool.
    """

    name: str
    args: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        return {"name": self.name, "args": self.args}


@dataclass
class CapturedRun:
    """
    A single capture run for a case, containing tool calls.

    Attributes:
        tool_calls: List of tool calls made by the model in this run.
    """

    tool_calls: list[CapturedToolCall] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        return {"tool_calls": [tc.to_dict() for tc in self.tool_calls]}


@dataclass
class CapturedCase:
    """
    Result of running a single case in capture mode.

    Attributes:
        case_name: The name of the evaluation case.
        user_message: The user message that triggered the tool calls.
        tool_calls: List of tool calls made by the model.
        system_message: The system message (included if include_context is True).
        additional_messages: Additional messages (included if include_context is True).
        track_name: The track name for comparative captures (None for regular cases).
        runs: Optional list of runs (populated when num_runs > 1).
    """

    case_name: str
    user_message: str
    tool_calls: list[CapturedToolCall] = field(default_factory=list)
    system_message: str | None = None
    additional_messages: list[dict[str, Any]] | None = None
    track_name: str | None = None
    runs: list[CapturedRun] = field(default_factory=list)

    @staticmethod
    def _try_parse_json(value: str) -> Any:
        """Try to parse a JSON string, returning the original string if parsing fails."""
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return value

    @staticmethod
    def _normalize_messages(messages: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """
        Normalize additional_messages by parsing JSON strings into proper objects.

        OpenAI returns:
        - Tool call arguments as JSON strings in assistant messages
        - Tool response content as JSON strings in tool messages

        For cleaner output, we parse these into proper objects.
        """
        normalized = []
        for msg in messages:
            msg_copy = dict(msg)

            # Parse tool call arguments in assistant messages
            if "tool_calls" in msg_copy and isinstance(msg_copy["tool_calls"], list):
                normalized_tool_calls = []
                for tc in msg_copy["tool_calls"]:
                    tc_copy = dict(tc)
                    if "function" in tc_copy and isinstance(tc_copy["function"], dict):
                        func = dict(tc_copy["function"])
                        if "arguments" in func and isinstance(func["arguments"], str):
                            func["arguments"] = CapturedCase._try_parse_json(func["arguments"])
                        tc_copy["function"] = func
                    normalized_tool_calls.append(tc_copy)
                msg_copy["tool_calls"] = normalized_tool_calls

            # Parse content in tool response messages
            if msg_copy.get("role") == "tool" and isinstance(msg_copy.get("content"), str):
                msg_copy["content"] = CapturedCase._try_parse_json(msg_copy["content"])

            normalized.append(msg_copy)
        return normalized

    def to_dict(self, include_context: bool = False) -> dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        result: dict[str, Any] = {
            "case_name": self.case_name,
            "user_message": self.user_message,
            "tool_calls": [tc.to_dict() for tc in self.tool_calls],
        }
        if self.runs:
            result["runs"] = [run.to_dict() for run in self.runs]
        if self.track_name:
            result["track_name"] = self.track_name
        if include_context:
            result["system_message"] = self.system_message
            # Normalize additional_messages to parse JSON string arguments
            raw_messages = self.additional_messages or []
            result["additional_messages"] = self._normalize_messages(raw_messages)
        return result


@dataclass
class CaptureResult:
    """
    Result of running an EvalSuite in capture mode.

    Attributes:
        suite_name: The name of the evaluation suite.
        model: The model used for capture.
        provider: The provider used (openai, anthropic).
        captured_cases: List of captured cases with tool calls.
    """

    suite_name: str
    model: str
    provider: str
    captured_cases: list[CapturedCase] = field(default_factory=list)

    def to_dict(self, include_context: bool = False) -> dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        return {
            "suite_name": self.suite_name,
            "model": self.model,
            "provider": self.provider,
            "captured_cases": [c.to_dict(include_context) for c in self.captured_cases],
        }

    def to_json(self, include_context: bool = False, indent: int = 2) -> str:
        """Convert to JSON string."""
        return json.dumps(self.to_dict(include_context), indent=indent)

    def write_to_file(self, file_path: str, include_context: bool = False, indent: int = 2) -> None:
        """Write capture results to a JSON file."""
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(self.to_json(include_context, indent))


# --- Helper functions for running capture mode ---


async def _capture_with_openai(
    suite: EvalSuite,
    api_key: str,
    model: str,
    include_context: bool = False,
    num_runs: int = 1,
    seed: str | int | None = "constant",
) -> CaptureResult:
    """Run capture mode with OpenAI client."""
    async with AsyncOpenAI(api_key=api_key) as client:
        return await suite.capture(
            client,
            model,
            provider="openai",
            include_context=include_context,
            num_runs=num_runs,
            seed=seed,
        )


async def _capture_with_anthropic(
    suite: EvalSuite,
    api_key: str,
    model: str,
    include_context: bool = False,
    num_runs: int = 1,
    seed: str | int | None = "constant",
) -> CaptureResult:
    """Run capture mode with Anthropic client."""
    try:
        from anthropic import AsyncAnthropic
    except ImportError as e:
        raise ImportError(
            "The 'anthropic' package is required for Anthropic provider. "
            "Install it with: pip install anthropic"
        ) from e

    async with AsyncAnthropic(api_key=api_key) as client:
        return await suite.capture(
            client,
            model,
            provider="anthropic",
            include_context=include_context,
            num_runs=num_runs,
            seed=seed,
        )



"""Tests for MCP Server implementation."""

import asyncio
import contextlib
from typing import Annotated
from unittest.mock import AsyncMock, Mock

import pytest
from arcade_core.auth import OAuth2
from arcade_core.catalog import MaterializedTool, ToolMeta, create_func_models
from arcade_core.errors import ToolRuntimeError
from arcade_core.schema import (
    InputParameter,
    OAuth2Requirement,
    ToolAuthRequirement,
    ToolContext,
    ToolDefinition,
    ToolInput,
    ToolkitDefinition,
    ToolOutput,
    ToolRequirements,
    ToolSecretRequirement,
    ValueSchema,
)
from arcade_mcp_server import tool
from arcade_mcp_server.middleware import Middleware
from arcade_mcp_server.server import MCPServer
from arcade_mcp_server.session import InitializationState
from arcade_mcp_server.types import (
    CallToolRequest,
    CallToolResult,
    InitializeRequest,
    InitializeResult,
    JSONRPCError,
    JSONRPCResponse,
    ListToolsRequest,
    ListToolsResult,
    PingRequest,
)


class TestMCPServer:
    """Test MCPServer class."""

    def test_server_initialization(self, tool_catalog, mcp_settings):
        """Test server initialization with various configurations."""
        # Basic initialization
        server = MCPServer(
            catalog=tool_catalog,
            name="Test Server",
            version="1.9.0",
            settings=mcp_settings,
        )

        assert server.name == "Test Server"
        assert server.version == "1.9.0"
        assert server.title == "Test Server"
        assert server.settings == mcp_settings

        # With custom title and instructions
        server2 = MCPServer(
            catalog=tool_catalog,
            name="Test Server",
            version="1.0.0",
            title="Custom Title",
            instructions="Custom instructions",
        )

        assert server2.title == "Custom Title"
        assert server2.instructions == "Custom instructions"

    def test_server_initialization_with_settings_defaults(self, tool_catalog):
        """Test server initialization uses settings when parameters not provided."""
        from arcade_mcp_server.settings import MCPSettings, ServerSettings

        settings = MCPSettings(
            server=ServerSettings(
                name="SettingsName",
                version="2.0.0",
                title="SettingsTitle",
                instructions="Settings instructions",
            )
        )

        # Initialize without name/version - should use settings
        server = MCPServer(catalog=tool_catalog, settings=settings)

        assert server.name == "SettingsName"
        assert server.version == "2.0.0"
        assert server.title == "SettingsTitle"
        assert server.instructions == "Settings instructions"

    def test_server_initialization_parameters_override_settings(self, tool_catalog):
        """Test server initialization parameters override settings."""
        from arcade_mcp_server.settings import MCPSettings, ServerSettings

        settings = MCPSettings(
            server=ServerSettings(
                name="SettingsName",
                version="2.0.0",
                title="SettingsTitle",
                instructions="Settings instructions",
            )
        )

        # Initialize with explicit parameters (should override settings)
        server = MCPServer(
            catalog=tool_catalog,
            name="ParamName",
            version="3.0.0",
            title="ParamTitle",
            instructions="Param instructions",
            settings=settings,
        )

        assert server.name == "ParamName"
        assert server.version == "3.0.0"
        assert server.title == "ParamTitle"
        assert server.instructions == "Param instructions"

    def test_server_initialization_title_fallback_logic(self, tool_catalog):
        """Test server initialization title fallback logic."""
        from arcade_mcp_server.settings import MCPSettings, ServerSettings

        # Test 1: Title parameter provided (should be used)
        server1 = MCPServer(
            catalog=tool_catalog,
            name="TestServer",
            title="ExplicitTitle",
        )
        assert server1.title == "ExplicitTitle"

        # Test 2: No title parameter but settings has non-default title
        settings2 = MCPSettings(
            server=ServerSettings(
                name="SettingsServer",
                title="CustomSettingsTitle",
            )
        )
        server2 = MCPServer(catalog=tool_catalog, settings=settings2)
        assert server2.title == "CustomSettingsTitle"

        # Test 3: No title parameter, settings has default title (should use name)
        settings3 = MCPSettings(
            server=ServerSettings(
                name="SettingsServer",
                title="ArcadeMCP",  # Default value
            )
        )
        server3 = MCPServer(catalog=tool_catalog, settings=settings3)
        assert server3.title == "SettingsServer"

        # Test 4: No title parameter, no settings title (should use name)
        settings4 = MCPSettings(
            server=ServerSettings(
                name="SettingsServer",
                title=None,
            )
        )
        server4 = MCPServer(catalog=tool_catalog, settings=settings4)
        assert server4.title == "SettingsServer"

    def test_server_initialization_instructions_fallback(self, tool_catalog):
        """Test server initialization instructions fallback logic."""
        from arcade_mcp_server.settings import MCPSettings, ServerSettings

        # Test 1: Instructions parameter provided (should be used)
        server1 = MCPServer(
            catalog=tool_catalog,
            instructions="Explicit instructions",
        )
        assert server1.instructions == "Explicit instructions"

        # Test 2: No instructions parameter (should use settings)
        settings2 = MCPSettings(
            server=ServerSettings(
                instructions="Settings instructions",
            )
        )
        server2 = MCPServer(catalog=tool_catalog, settings=settings2)
        assert server2.instructions == "Settings instructions"

        # Test 3: No instructions parameter, no settings (should use default)
        settings3 = MCPSettings(
            server=ServerSettings(
                instructions=None,
            )
        )
        server3 = MCPServer(catalog=tool_catalog, settings=settings3)
        assert "available tools" in server3.instructions.lower()

    def test_handler_registration(self, tool_catalog):
        """Test that all required handlers are registered."""
        server = MCPServer(catalog=tool_catalog)

        expected_handlers = [
            "ping",
            "initialize",
            "tools/list",
            "tools/call",
            "resources/list",
            "resources/templates/list",
            "resources/read",
            "prompts/list",
            "prompts/get",
            "logging/setLevel",
        ]

        for method in expected_handlers:
            assert method in server._handlers
            assert callable(server._handlers[method])

    @pytest.mark.asyncio
    async def test_server_lifecycle(self, tool_catalog, mcp_settings):
        """Test server startup and shutdown."""
        server = MCPServer(
            catalog=tool_catalog,
            settings=mcp_settings,
        )

        # Start server
        await server.start()

        # Stop server
        await server.stop()

    @pytest.mark.asyncio
    async def test_handle_ping(self, mcp_server):
        """Test ping request handling."""
        message = PingRequest(jsonrpc="2.0", id=1, method="ping")

        response = await mcp_server._handle_ping(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 1
        assert response.result == {}

    @pytest.mark.asyncio
    async def test_handle_initialize(self, mcp_server):
        """Test initialize request handling."""
        message = InitializeRequest(
            jsonrpc="2.0",
            id=1,
            method="initialize",
            params={
                "protocolVersion": "2024-11-05",
                "capabilities": {},
                "clientInfo": {"name": "test-client", "version": "1.0.0"},
            },
        )

        # Create mock session
        session = Mock()
        session.set_client_params = Mock()

        response = await mcp_server._handle_initialize(message, session=session)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 1
        assert isinstance(response.result, InitializeResult)
        assert response.result.protocolVersion is not None
        assert response.result.serverInfo.name == mcp_server.name
        assert response.result.serverInfo.version == mcp_server.version

        # Check session was updated
        session.set_client_params.assert_called_once()

    @pytest.mark.asyncio
    async def test_handle_list_tools(self, mcp_server):
        """Test list tools request handling."""
        message = ListToolsRequest(jsonrpc="2.0", id=2, method="tools/list", params={})

        response = await mcp_server._handle_list_tools(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 2
        assert isinstance(response.result, ListToolsResult)
        assert len(response.result.tools) > 0

    @pytest.mark.asyncio
    async def test_handle_call_tool(self, mcp_server):
        """Test tool call request handling."""
        message = CallToolRequest(
            jsonrpc="2.0",
            id=3,
            method="tools/call",
            params={"name": "TestToolkit.test_tool", "arguments": {"text": "Hello"}},
        )

        response = await mcp_server._handle_call_tool(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 3
        assert isinstance(response.result, CallToolResult)
        assert response.result.structuredContent is not None
        assert "result" in response.result.structuredContent
        assert "Echo: Hello" in response.result.structuredContent["result"]

    @pytest.mark.asyncio
    async def test_handle_call_tool_with_requires_auth(self, mcp_server):
        """Test tool call request handling with authorization."""

        # Mock arcade client so the server thinks API key is configured
        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        mock_auth_response = Mock()
        mock_auth_response.status = "pending"
        mock_auth_response.url = "https://example.com/auth"

        # Patch the _check_authorization method to return a tool that has unsatisfied authorization
        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        message = CallToolRequest(
            jsonrpc="2.0",
            id=3,
            method="tools/call",
            params={"name": "TestToolkit.sample_tool_with_auth", "arguments": {"text": "Hello"}},
        )

        response = await mcp_server._handle_call_tool(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 3
        assert isinstance(response.result, CallToolResult)
        assert response.result.structuredContent is not None
        assert "authorization_url" in response.result.structuredContent
        assert response.result.structuredContent["authorization_url"] == "https://example.com/auth"
        assert "message" in response.result.structuredContent
        assert "Authorization required" in response.result.structuredContent["message"]
        assert "needs your permission" in response.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_handle_call_tool_with_requires_auth_no_api_key(self, mcp_server):
        """Test tool call request handling with authorization when no Arcade API key is configured."""

        # Ensure no arcade client is configured
        mcp_server.arcade = None

        message = CallToolRequest(
            jsonrpc="2.0",
            id=3,
            method="tools/call",
            params={"name": "TestToolkit.sample_tool_with_auth", "arguments": {"text": "Hello"}},
        )

        response = await mcp_server._handle_call_tool(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.id == 3
        assert isinstance(response.result, CallToolResult)
        assert response.result.structuredContent is not None
        assert "message" in response.result.structuredContent
        assert "Missing Arcade API key" in response.result.structuredContent["message"]
        assert "requires authorization" in response.result.structuredContent["message"]
        assert "arcade login" in response.result.structuredContent["message"]
        assert "ARCADE_API_KEY" in response.result.structuredContent["message"]
        assert "ARCADE_API_KEY" in response.result.structuredContent["llm_instructions"]

    @pytest.mark.asyncio
    async def test_handle_call_tool_not_found(self, mcp_server):
        """Test calling a non-existent tool."""
        message = CallToolRequest(
            jsonrpc="2.0",
            id=3,
            method="tools/call",
            params={"name": "NonExistent.tool", "arguments": {}},
        )

        response = await mcp_server._handle_call_tool(message)

        assert isinstance(response, JSONRPCResponse)
        assert response.result.isError
        assert "error" in response.result.structuredContent
        assert "Unknown tool" in response.result.structuredContent["error"]

    @pytest.mark.asyncio
    async def test_handle_message_routing(self, mcp_server, initialized_server_session):
        """Test message routing to appropriate handlers."""
        # Test valid method
        message = {"jsonrpc": "2.0", "id": 1, "method": "ping"}

        response = await mcp_server.handle_message(message, session=initialized_server_session)

        assert response is not None
        assert str(response.id) == "1"
        assert response.result == {}

        # Test invalid method
        message = {"jsonrpc": "2.0", "id": 2, "method": "invalid/method"}

        response = await mcp_server.handle_message(message, session=initialized_server_session)

        assert isinstance(response, JSONRPCError)
        assert response.error["code"] == -32601
        assert "Method not found" in response.error["message"]

    @pytest.mark.asyncio
    async def test_handle_message_invalid_format(self, mcp_server):
        """Test handling of invalid message formats."""
        # Non-dict message
        response = await mcp_server.handle_message("invalid", session=None)

        assert isinstance(response, JSONRPCError)
        assert response.error["code"] == -32600
        assert "Invalid request" in response.error["message"]

    @pytest.mark.asyncio
    async def test_initialization_state_enforcement(self, mcp_server):
        """Test that non-initialize methods are blocked before initialization."""
        # Create uninitialized session
        session = Mock()
        session.initialization_state = InitializationState.NOT_INITIALIZED

        # Try to call tools/list before initialization
        message = {"jsonrpc": "2.0", "id": 1, "method": "tools/list"}

        response = await mcp_server.handle_message(message, session=session)

        assert isinstance(response, JSONRPCError)
        assert response.error["code"] == -32600
        assert "Not initialized" in response.error["message"]
        assert "cannot be processed before the session is initialized" in response.error["message"]

    @pytest.mark.asyncio
    async def test_notification_handling(self, mcp_server):
        """Test handling of notification messages."""
        session = Mock()
        session.mark_initialized = Mock()

        # Send initialized notification
        message = {"jsonrpc": "2.0", "method": "notifications/initialized"}

        response = await mcp_server.handle_message(message, session=session)

        # Notifications should not return a response
        assert response is None
        # Session should be marked as initialized
        session.mark_initialized.assert_called_once()

    @pytest.mark.asyncio
    async def test_middleware_chain(self, tool_catalog, mcp_settings):
        """Test middleware chain execution."""
        # Create a test middleware
        test_middleware_called = False

        class TestMiddleware(Middleware):
            async def __call__(self, context, call_next):
                nonlocal test_middleware_called
                test_middleware_called = True
                # Modify context
                context.metadata["test"] = "value"
                return await call_next(context)

        # Create server with middleware
        server = MCPServer(
            catalog=tool_catalog,
            settings=mcp_settings,
            middleware=[TestMiddleware()],
        )
        await server.start()

        # Send a message
        message = {"jsonrpc": "2.0", "id": 1, "method": "ping"}

        response = await server.handle_message(message)

        # Middleware should have been called
        assert test_middleware_called
        assert response is not None

    @pytest.mark.asyncio
    async def test_error_handling_middleware(self, mcp_server):
        """Test that error handling middleware catches exceptions."""

        # Mock a handler to raise an exception
        async def failing_handler(*args, **kwargs):
            raise Exception("Test error")

        mcp_server._handlers["test/fail"] = failing_handler

        message = {"jsonrpc": "2.0", "id": 1, "method": "test/fail"}

        response = await mcp_server.handle_message(message)

        assert isinstance(response, JSONRPCError)
        assert response.error["code"] == -32603
        # Error details should be masked in production
        if mcp_server.settings.middleware.mask_error_details:
            assert response.error["message"] == "Internal error"
        else:
            assert "Test error" in response.error["message"]

    @pytest.mark.asyncio
    async def test_session_management(self, mcp_server):
        """Test session creation and cleanup."""

        # Create a mock read stream that waits
        async def mock_stream():
            try:
                while True:
                    await asyncio.sleep(1)  # Keep the session alive
                    yield None  # Yield nothing
            except asyncio.CancelledError:
                pass

        mock_read_stream = mock_stream()
        mock_write_stream = AsyncMock()

        # Track sessions
        initial_sessions = len(mcp_server._sessions)

        # Create a new connection
        session_task = asyncio.create_task(
            mcp_server.run_connection(mock_read_stream, mock_write_stream)
        )

        # Give it time to register
        await asyncio.sleep(0.1)

        # Should have one more session
        assert len(mcp_server._sessions) == initial_sessions + 1

        # Cancel the session
        session_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await session_task

        # Give it time to clean up
        await asyncio.sleep(0.1)

        # Session should be cleaned up
        assert len(mcp_server._sessions) == initial_sessions

    @pytest.mark.asyncio
    async def test_authorization_check(self, mcp_server):
        """Test tool authorization checking."""

        # Ensure the arcade client is not configured in the case that the test environment
        # unintentionally has the ARCADE_API_KEY set
        mcp_server.arcade = None

        tool = Mock()
        tool.definition.requirements.authorization = ToolAuthRequirement(
            provider_type="oauth2", provider_id="test-provider"
        )

        # Without arcade client configured
        with pytest.raises(Exception) as exc_info:
            await mcp_server._check_authorization(tool)

        assert "Authorization check called without Arcade API key configured" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_check_tool_requirements_no_requirements(self, mcp_server, materialized_tool):
        """Test tool requirements checking when tool has no requirements."""

        # Create a tool with no requirements
        tool = materialized_tool
        tool.definition.requirements = None

        tool_context = ToolContext()
        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.test_tool", "arguments": {"text": "Hello"}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.test_tool"
        )

        # Should return None when no requirements because this means the tool can be executed
        assert result is None

    @pytest.mark.asyncio
    async def test_check_tool_requirements_auth_no_arcade_client(self, mcp_server):
        """Test tool requirements checking when tool requires auth but no Arcade client configured."""

        # Ensure no arcade client is configured
        mcp_server.arcade = None

        # Create a tool that requires authorization
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            )
        )

        tool_context = ToolContext()
        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.auth_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.auth_tool"
        )

        # Should return error response
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "Missing Arcade API key" in result.result.structuredContent["message"]
        assert "requires authorization" in result.result.structuredContent["message"]
        assert "ARCADE_API_KEY" in result.result.structuredContent["message"]
        assert "ARCADE_API_KEY" in result.result.structuredContent["llm_instructions"]

    @pytest.mark.asyncio
    async def test_check_tool_requirements_auth_pending(self, mcp_server):
        """Test tool requirements checking when authorization is pending."""

        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        # Create a tool that requires authorization
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            )
        )

        mock_auth_response = Mock()
        mock_auth_response.status = "pending"
        mock_auth_response.url = "https://example.com/auth"

        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        tool_context = ToolContext()
        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.auth_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.auth_tool"
        )

        # Should return error response with authorization URL
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "authorization_url" in result.result.structuredContent
        assert result.result.structuredContent["authorization_url"] == "https://example.com/auth"
        assert "Authorization required" in result.result.structuredContent["message"]
        assert "needs your permission" in result.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_check_tool_requirements_auth_completed(self, mcp_server):
        """Test tool requirements checking when authorization is completed."""

        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        # Create a tool that requires authorization
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            )
        )

        # Mock authorization response as completed
        mock_auth_response = Mock()
        mock_auth_response.status = "completed"
        mock_auth_response.context = Mock()
        mock_auth_response.context.token = "test-token"
        mock_auth_response.context.user_info = {"user_id": "test-user"}

        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        tool_context = ToolContext()
        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.auth_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.auth_tool"
        )

        # Should return None (no error) and set authorization context
        assert result is None
        assert tool_context.authorization is not None
        assert tool_context.authorization.token == "test-token"
        assert tool_context.authorization.user_info == {"user_id": "test-user"}

    @pytest.mark.asyncio
    async def test_check_tool_requirements_auth_error(self, mcp_server):
        """Test tool requirements checking when authorization fails."""

        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        # Create a tool that requires authorization
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            )
        )

        # Mock authorization to raise an error
        mcp_server._check_authorization = AsyncMock(side_effect=ToolRuntimeError("Auth failed"))

        tool_context = ToolContext()
        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.auth_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.auth_tool"
        )

        # Should return error response
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "Authorization error" in result.result.structuredContent["message"]
        assert "failed to authorize" in result.result.structuredContent["message"]
        assert "Auth failed" in result.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_check_tool_requirements_secrets_missing(self, mcp_server):
        """Test tool requirements checking when required secrets are missing."""

        # Create a tool that requires secrets
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            secrets=[
                ToolSecretRequirement(key="API_KEY"),
                ToolSecretRequirement(key="DATABASE_URL"),
            ]
        )

        # Mock tool context to raise ValueError for missing secrets
        tool_context = Mock(spec=ToolContext)
        tool_context.get_secret = Mock(side_effect=ValueError("Secret not found"))

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.secret_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.secret_tool"
        )

        # Should return error response
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "Missing secret" in result.result.structuredContent["message"]
        assert "API_KEY, DATABASE_URL" in result.result.structuredContent["message"]
        assert ".env file" in result.result.structuredContent["message"]
        assert ".env file" in result.result.structuredContent["llm_instructions"]

    @pytest.mark.asyncio
    async def test_check_tool_requirements_secrets_partial_missing(self, mcp_server):
        """Test tool requirements checking when some required secrets are missing."""

        # Create a tool that requires secrets
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            secrets=[
                ToolSecretRequirement(key="API_KEY"),
                ToolSecretRequirement(key="DATABASE_URL"),
            ]
        )

        # Mock tool context to return a strict subset of the required secrets
        tool_context = Mock(spec=ToolContext)

        def mock_get_secret(key):
            if key == "API_KEY":
                return "test-api-key"
            else:
                raise ValueError("Secret not found")

        tool_context.get_secret = Mock(side_effect=mock_get_secret)

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.secret_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.secret_tool"
        )

        # Should return error response for missing DATABASE_URL
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "DATABASE_URL" in result.result.structuredContent["message"]
        assert "API_KEY" not in result.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_check_tool_requirements_secrets_available(self, mcp_server):
        """Test tool requirements checking when all required secrets are available."""

        # Create a tool that requires secrets
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            secrets=[
                ToolSecretRequirement(key="API_KEY"),
                ToolSecretRequirement(key="DATABASE_URL"),
            ]
        )

        # Mock tool context to return all secrets
        tool_context = Mock(spec=ToolContext)

        def mock_get_secret(key):
            return f"test-{key.lower()}-value"

        tool_context.get_secret = Mock(side_effect=mock_get_secret)

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.secret_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.secret_tool"
        )

        # Should return None (no error) when all secrets are available
        assert result is None

    @pytest.mark.asyncio
    async def test_check_tool_requirements_combined_auth_and_secrets(self, mcp_server):
        """Test tool requirements checking with both auth and secrets requirements."""

        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        # Create a tool that requires both auth and secrets
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            ),
            secrets=[
                ToolSecretRequirement(key="API_KEY"),
            ],
        )

        # Mock successful authorization
        mock_auth_response = Mock()
        mock_auth_response.status = "completed"
        mock_auth_response.context = Mock()
        mock_auth_response.context.token = "test-token"
        mock_auth_response.context.user_info = {"user_id": "test-user"}

        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        tool_context = ToolContext()
        tool_context.set_secret("API_KEY", "test-api-key")

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.combined_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.combined_tool"
        )

        # Should return None (no error) when both requirements are satisfied
        assert result is None
        # Authorization context should be set
        assert tool_context.authorization is not None

    @pytest.mark.asyncio
    async def test_check_tool_requirements_combined_auth_fails_first(self, mcp_server):
        """Test tool requirements checking when auth fails before secrets are checked."""

        mock_arcade = Mock()
        mcp_server.arcade = mock_arcade

        # Create a tool that requires both auth and secrets
        tool = Mock()
        tool.definition.requirements = ToolRequirements(
            authorization=ToolAuthRequirement(
                provider_type="oauth2",
                provider_id="test-provider",
            ),
            secrets=[
                ToolSecretRequirement(key="API_KEY"),
            ],
        )

        # Mock authorization as pending (should fail before secrets check)
        mock_auth_response = Mock()
        mock_auth_response.status = "pending"
        mock_auth_response.url = "https://example.com/auth"

        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        # Create real tool context (secrets check shouldn't be reached)
        tool_context = ToolContext()
        tool_context.set_secret("API_KEY", "test-api-key")

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.combined_tool", "arguments": {}},
        )

        result = await mcp_server._check_tool_requirements(
            tool, tool_context, message, "TestToolkit.combined_tool"
        )

        # Should return auth error (auth is checked first)
        assert isinstance(result, JSONRPCResponse)
        assert isinstance(result.result, CallToolResult)
        assert result.result.isError is True
        assert "authorization_url" in result.result.structuredContent

    @pytest.mark.asyncio
    async def test_http_transport_blocks_tool_with_auth(
        self, mcp_server, materialized_tool_with_auth
    ):
        """Test that HTTP transport blocks tools requiring oauth."""
        # Create a mock session with HTTP transport
        session = Mock()
        session.init_options = {"transport_type": "http"}

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={
                "name": "TestToolkit.sample_tool_with_auth",
                "arguments": {"text": "test"},
            },
        )
        response = await mcp_server._handle_call_tool(message, session=session)

        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)
        assert response.result.isError is True
        assert "HTTP transport" in response.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_http_transport_blocks_tool_with_secrets(self, mcp_server):
        """Test that HTTP transport blocks tools requiring secrets."""
        from arcade_core.schema import ToolSecretRequirement

        tool_def = ToolDefinition(
            name="secret_tool",
            fully_qualified_name="TestToolkit.secret_tool",
            description="A tool requiring secrets",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="text",
                        required=True,
                        description="Input text",
                        value_schema=ValueSchema(val_type="string"),
                    )
                ]
            ),
            output=ToolOutput(
                description="Tool output", value_schema=ValueSchema(val_type="string")
            ),
            requirements=ToolRequirements(
                secrets=[ToolSecretRequirement(key="API_KEY", description="API Key")]
            ),
        )

        @tool(requires_secrets=["SECRET_KEY"])
        def secret_tool_func(text: Annotated[str, "Input text"]) -> Annotated[str, "Secret text"]:
            """Secret tool function"""
            return "Secret"

        input_model, output_model = create_func_models(secret_tool_func)
        meta = ToolMeta(module=secret_tool_func.__module__, toolkit="TestToolkit")
        materialized_tool = MaterializedTool(
            tool=secret_tool_func,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )

        await mcp_server._tool_manager.add_tool(materialized_tool)

        # Create a mock session with HTTP transport
        session = Mock()
        session.init_options = {"transport_type": "http"}

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.secret_tool", "arguments": {"text": "test"}},
        )

        response = await mcp_server._handle_call_tool(message, session=session)

        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)
        assert response.result.isError is True
        assert "HTTP transport" in response.result.structuredContent["message"]
        assert "secrets" in response.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_http_transport_blocks_tool_with_both_auth_and_secrets(self, mcp_server):
        """Test that HTTP transport blocks tools requiring both auth and secrets."""
        from arcade_core.schema import ToolSecretRequirement

        # Create a tool with both auth and secret requirements
        tool_def = ToolDefinition(
            name="combined_tool",
            fully_qualified_name="TestToolkit.combined_tool",
            description="A tool requiring both auth and secrets",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="text",
                        required=True,
                        description="Input text",
                        value_schema=ValueSchema(val_type="string"),
                    )
                ]
            ),
            output=ToolOutput(
                description="Tool output", value_schema=ValueSchema(val_type="string")
            ),
            requirements=ToolRequirements(
                authorization=ToolAuthRequirement(
                    provider_type="oauth2",
                    provider_id="test-provider",
                    id="test-provider",
                    oauth2=OAuth2Requirement(scopes=["test.scope"]),
                ),
                secrets=[ToolSecretRequirement(key="API_KEY", description="API Key")],
            ),
        )

        @tool(
            requires_auth=OAuth2(id="test-provider", scopes=["test.scope"]),
            requires_secrets=["API_KEY"],
        )
        def combined_tool_func(
            text: Annotated[str, "Input text"],
        ) -> Annotated[str, "Combined text"]:
            """Combined tool function"""
            return f"Combined: {text}"

        input_model, output_model = create_func_models(combined_tool_func)
        meta = ToolMeta(module=combined_tool_func.__module__, toolkit="TestToolkit")
        materialized_tool = MaterializedTool(
            tool=combined_tool_func,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )

        await mcp_server._tool_manager.add_tool(materialized_tool)

        # Create a mock session with HTTP transport
        session = Mock()
        session.init_options = {"transport_type": "http"}

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.combined_tool", "arguments": {"text": "test"}},
        )

        response = await mcp_server._handle_call_tool(message, session=session)

        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)
        assert response.result.isError is True
        assert "Unsupported transport" in response.result.structuredContent["message"]
        assert "HTTP transport" in response.result.structuredContent["message"]
        assert "authorization" in response.result.structuredContent["message"]

    @pytest.mark.asyncio
    async def test_stdio_transport_allows_tool_with_auth(
        self, mcp_server, materialized_tool_with_auth
    ):
        """Test that stdio transport allows tools requiring authentication."""
        # Mock Arcade client
        mcp_server.arcade = Mock()
        mock_auth_response = Mock()
        mock_auth_response.status = "completed"
        mock_auth_response.context = Mock()
        mock_auth_response.context.token = "test-token"
        mock_auth_response.context.user_info = {}
        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        # Create a mock session with stdio transport
        session = Mock()
        session.init_options = {"transport_type": "stdio"}
        session.session_id = "test-session"

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={
                "name": "TestToolkit.sample_tool_with_auth",
                "arguments": {"text": "test"},
            },
        )

        response = await mcp_server._handle_call_tool(message, session=session)

        # Should succeed (isn't blocked by transport check)
        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)

        assert response.result.isError is False

    @pytest.mark.asyncio
    async def test_no_transport_type_allows_tool_with_auth(
        self, mcp_server, materialized_tool_with_auth
    ):
        """Test backwards compatibility: no transport_type specified allows tools."""
        # Mock Arcade client
        mcp_server.arcade = Mock()
        mock_auth_response = Mock()
        mock_auth_response.status = "completed"
        mock_auth_response.context = Mock()
        mock_auth_response.context.token = "test-token"
        mock_auth_response.context.user_info = {}
        mcp_server._check_authorization = AsyncMock(return_value=mock_auth_response)

        # Create a mock session without transport_type
        session = Mock()
        session.init_options = {}  # No transport_type
        session.session_id = "test-session"

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={
                "name": "TestToolkit.sample_tool_with_auth",
                "arguments": {"text": "test"},
            },
        )

        response = await mcp_server._handle_call_tool(message, session=session)

        # Should succeed (no transport restriction applies)
        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)
        assert response.result.isError is False

    @pytest.mark.asyncio
    async def test_http_transport_allows_tool_without_requirements(self, mcp_server):
        """Test that HTTP transport allows tools without auth/secret requirements."""
        # Create a mock session with HTTP transport
        session = Mock()
        session.init_options = {"transport_type": "http"}
        session.session_id = "test-session"

        message = CallToolRequest(
            jsonrpc="2.0",
            id=1,
            method="tools/call",
            params={"name": "TestToolkit.test_tool", "arguments": {"text": "test"}},
        )

        response = await mcp_server._handle_call_tool(message, session=session)

        assert isinstance(response, JSONRPCResponse)
        assert isinstance(response.result, CallToolResult)
        assert response.result.isError is False


class TestMissingSecretsWarnings:
    """Test startup warnings for missing tool secrets."""

    @pytest.mark.asyncio
    async def test_warns_missing_secrets_on_startup(self, tool_catalog, mcp_settings, caplog):
        """Test that missing secrets trigger warnings during server startup."""
        import logging

        # Create tool definition with secret requirements
        tool_def = ToolDefinition(
            name="fetch_data",
            fully_qualified_name="TestToolkit.fetch_data",
            description="Fetch data from API.",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="query",
                        required=True,
                        description="Search query",
                        value_schema=ValueSchema(val_type="string"),
                    )
                ]
            ),
            output=ToolOutput(description="Result", value_schema=ValueSchema(val_type="string")),
            requirements=ToolRequirements(
                secrets=[
                    ToolSecretRequirement(key="API_KEY", description="API Key"),
                    ToolSecretRequirement(key="SECRET_TOKEN", description="Secret Token"),
                ]
            ),
        )

        @tool
        def fetch_data(query: Annotated[str, "Search query"]) -> Annotated[str, "Result"]:
            """Fetch data from API."""
            return f"Data for {query}"

        # Add tool to catalog

        input_model, output_model = create_func_models(fetch_data)
        meta = ToolMeta(module=fetch_data.__module__, toolkit="TestToolkit")
        materialized = MaterializedTool(
            tool=fetch_data,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )
        tool_catalog._tools[tool_def.get_fully_qualified_name()] = materialized

        # Clear any existing secrets from environment
        import os

        old_api_key = os.environ.pop("API_KEY", None)
        old_secret_token = os.environ.pop("SECRET_TOKEN", None)

        try:
            # Ensure worker routes are disabled (no ARCADE_WORKER_SECRET)
            mcp_settings.arcade.server_secret = None

            # Create and start server
            with caplog.at_level(logging.WARNING):
                server = MCPServer(
                    catalog=tool_catalog,
                    name="Test Server",
                    version="1.0.0",
                    settings=mcp_settings,
                )
                await server.start()

                # Check for warning message
                warning_messages = [
                    rec.message for rec in caplog.records if rec.levelno == logging.WARNING
                ]

                # Should have a warning about missing secrets
                assert any("fetch_data" in msg and "API_KEY" in msg for msg in warning_messages), (
                    f"Expected warning about missing API_KEY for fetch_data. Got: {warning_messages}"
                )
                assert any(
                    "fetch_data" in msg and "SECRET_TOKEN" in msg for msg in warning_messages
                ), (
                    f"Expected warning about missing SECRET_TOKEN for fetch_data. Got: {warning_messages}"
                )

                await server.stop()
        finally:
            # Restore environment
            if old_api_key is not None:
                os.environ["API_KEY"] = old_api_key
            if old_secret_token is not None:
                os.environ["SECRET_TOKEN"] = old_secret_token

    @pytest.mark.asyncio
    async def test_no_warning_when_secrets_present(self, tool_catalog, mcp_settings, caplog):
        """Test that no warnings are shown when secrets are available."""
        import logging

        # Create tool definition with secret requirements
        tool_def = ToolDefinition(
            name="secure_tool",
            fully_qualified_name="TestToolkit.secure_tool",
            description="Secure tool.",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="data",
                        required=True,
                        description="Data",
                        value_schema=ValueSchema(val_type="string"),
                    )
                ]
            ),
            output=ToolOutput(description="Result", value_schema=ValueSchema(val_type="string")),
            requirements=ToolRequirements(
                secrets=[ToolSecretRequirement(key="PRESENT_KEY", description="Present Key")]
            ),
        )

        @tool
        def secure_tool(data: Annotated[str, "Data"]) -> Annotated[str, "Result"]:
            """Secure tool."""
            return f"Processed {data}"

        # Add tool to catalog

        input_model, output_model = create_func_models(secure_tool)
        meta = ToolMeta(module=secure_tool.__module__, toolkit="TestToolkit")
        materialized = MaterializedTool(
            tool=secure_tool,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )
        tool_catalog._tools[tool_def.get_fully_qualified_name()] = materialized

        # Set the secret in environment
        import os

        old_value = os.environ.get("PRESENT_KEY")
        os.environ["PRESENT_KEY"] = "test-value"

        try:
            # Ensure worker routes are disabled
            mcp_settings.arcade.server_secret = None

            # Create and start server
            with caplog.at_level(logging.WARNING):
                server = MCPServer(
                    catalog=tool_catalog,
                    name="Test Server",
                    version="1.0.0",
                    settings=mcp_settings,
                )
                await server.start()

                # Check that no warning is logged for this tool
                warning_messages = [
                    rec.message for rec in caplog.records if rec.levelno == logging.WARNING
                ]
                assert not any(
                    "secure_tool" in msg and "PRESENT_KEY" in msg for msg in warning_messages
                ), f"Should not warn about PRESENT_KEY when it's set. Got: {warning_messages}"

                await server.stop()
        finally:
            # Restore environment
            if old_value is not None:
                os.environ["PRESENT_KEY"] = old_value
            else:
                os.environ.pop("PRESENT_KEY", None)

    @pytest.mark.asyncio
    async def test_no_warning_when_worker_routes_enabled(self, tool_catalog, mcp_settings, caplog):
        """Test that warnings are skipped when worker routes are enabled."""
        import logging

        # Create tool definition with secret requirements
        tool_def = ToolDefinition(
            name="worker_tool",
            fully_qualified_name="TestToolkit.worker_tool",
            description="Worker tool.",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="param",
                        required=True,
                        description="Param",
                        value_schema=ValueSchema(val_type="string"),
                    )
                ]
            ),
            output=ToolOutput(description="Result", value_schema=ValueSchema(val_type="string")),
            requirements=ToolRequirements(
                secrets=[ToolSecretRequirement(key="WORKER_API_KEY", description="Worker API Key")]
            ),
        )

        @tool
        def worker_tool(param: Annotated[str, "Param"]) -> Annotated[str, "Result"]:
            """Worker tool."""
            return f"Result: {param}"

        # Add tool to catalog

        input_model, output_model = create_func_models(worker_tool)
        meta = ToolMeta(module=worker_tool.__module__, toolkit="TestToolkit")
        materialized = MaterializedTool(
            tool=worker_tool,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )
        tool_catalog._tools[tool_def.get_fully_qualified_name()] = materialized

        # Clear the secret from environment
        import os

        old_value = os.environ.pop("WORKER_API_KEY", None)

        try:
            # Enable worker routes by setting ARCADE_WORKER_SECRET
            mcp_settings.arcade.server_secret = "test-worker-secret"

            # Create and start server
            with caplog.at_level(logging.WARNING):
                server = MCPServer(
                    catalog=tool_catalog,
                    name="Test Server",
                    version="1.0.0",
                    settings=mcp_settings,
                )
                await server.start()

                # Check that no warning is logged (worker routes are enabled)
                warning_messages = [
                    rec.message for rec in caplog.records if rec.levelno == logging.WARNING
                ]
                assert not any(
                    "worker_tool" in msg and "WORKER_API_KEY" in msg for msg in warning_messages
                ), f"Should not warn when worker routes are enabled. Got: {warning_messages}"

                await server.stop()
        finally:
            # Restore environment
            if old_value is not None:
                os.environ["WORKER_API_KEY"] = old_value

    @pytest.mark.asyncio
    async def test_warning_format(self, tool_catalog, mcp_settings, caplog):
        """Test that warnings use the expected format."""
        import logging

        # Create tool definition with secret requirement
        tool_def = ToolDefinition(
            name="format_test_tool",
            fully_qualified_name="TestToolkit.format_test_tool",
            description="Format test tool.",
            toolkit=ToolkitDefinition(
                name="TestToolkit", description="Test toolkit", version="1.0.0"
            ),
            input=ToolInput(
                parameters=[
                    InputParameter(
                        name="x",
                        required=True,
                        description="Input",
                        value_schema=ValueSchema(val_type="integer"),
                    )
                ]
            ),
            output=ToolOutput(description="Output", value_schema=ValueSchema(val_type="integer")),
            requirements=ToolRequirements(
                secrets=[
                    ToolSecretRequirement(key="FORMAT_TEST_KEY", description="Format Test Key")
                ]
            ),
        )

        @tool
        def format_test_tool(x: Annotated[int, "Input"]) -> Annotated[int, "Output"]:
            """Format test tool."""
            return x * 2

        # Add tool to catalog
        input_model, output_model = create_func_models(format_test_tool)
        meta = ToolMeta(module=format_test_tool.__module__, toolkit="TestToolkit")
        materialized = MaterializedTool(
            tool=format_test_tool,
            definition=tool_def,
            meta=meta,
            input_model=input_model,
            output_model=output_model,
        )
        tool_catalog._tools[tool_def.get_fully_qualified_name()] = materialized

        # Clear the secret from environment
        import os

        old_value = os.environ.pop("FORMAT_TEST_KEY", None)

        try:
            # Ensure worker routes are disabled
            mcp_settings.arcade.server_secret = None

            # Create and start server
            with caplog.at_level(logging.WARNING):
                server = MCPServer(
                    catalog=tool_catalog,
                    name="Test Server",
                    version="1.0.0",
                    settings=mcp_settings,
                )
                await server.start()

                # Check warning format matches specification
                warning_messages = [
                    rec.message for rec in caplog.records if rec.levelno == logging.WARNING
                ]

                # Find the warning for our tool
                matching_warnings = [msg for msg in warning_messages if "format_test_tool" in msg]
                assert len(matching_warnings) > 0, (
                    f"Expected warning for format_test_tool. Got: {warning_messages}"
                )

                warning = matching_warnings[0]
                # Check format: "⚠ Tool 'name' declares secret(s) 'KEY' which are not set"
                assert "Tool 'format_test_tool'" in warning
                assert "not set" in warning

                await server.stop()
        finally:
            # Restore environment
            if old_value is not None:
                os.environ["FORMAT_TEST_KEY"] = old_value



import os

# The path to the directory containing the Arcade configuration files. Typically ~/.arcade
ARCADE_CONFIG_PATH = os.path.join(os.path.expanduser(os.getenv("ARCADE_WORK_DIR", "~")), ".arcade")
# The path to the file containing the user's Arcade-related credentials (e.g., ARCADE_API_KEY).
CREDENTIALS_FILE_PATH = os.path.join(ARCADE_CONFIG_PATH, "credentials.yaml")

# Host defaults used by both the CLI and MCP server
PROD_COORDINATOR_HOST = "cloud.arcade.dev"
PROD_ENGINE_HOST = "api.arcade.dev"
LOCALHOST = "localhost"



"""Tests for Tool Manager implementation."""

import pytest
import pytest_asyncio
from arcade_mcp_server.exceptions import NotFoundError
from arcade_mcp_server.managers.tool import ToolManager
from arcade_mcp_server.types import MCPTool


class TestToolManager:
    """Test ToolManager class."""

    @pytest_asyncio.fixture
    async def tool_manager(self, materialized_tool):
        """Create a tool manager instance with one tool added."""
        manager = ToolManager()
        await manager.add_tool(materialized_tool)
        return manager

    def test_manager_initialization(self):
        """Test tool manager initialization."""
        manager = ToolManager()
        assert isinstance(manager, ToolManager)

    @pytest.mark.asyncio
    async def test_list_tools(self, tool_manager):
        """Test listing tools."""
        tools = await tool_manager.list_tools()

        assert isinstance(tools, list)
        assert all(isinstance(t, MCPTool) for t in tools)

        if tools:
            tool = tools[0]
            assert hasattr(tool, "name")
            assert hasattr(tool, "description")
            assert hasattr(tool, "inputSchema")

    @pytest.mark.asyncio
    async def test_get_tool(self, tool_manager, materialized_tool):
        """Test getting a specific tool."""
        # Get tool by name
        tool_name = materialized_tool.definition.fully_qualified_name
        tool = await tool_manager.get_tool(tool_name)
        assert tool.definition.fully_qualified_name == tool_name

        # Try to get non-existent tool
        with pytest.raises(NotFoundError):
            await tool_manager.get_tool("NonExistent_tool")

    @pytest.mark.asyncio
    async def test_remove_tool(self, tool_manager, materialized_tool):
        """Test removing tools."""
        name = materialized_tool.definition.fully_qualified_name
        _ = await tool_manager.get_tool(name)

        removed = await tool_manager.remove_tool(name)
        assert removed.definition.fully_qualified_name == name

        with pytest.raises(NotFoundError):
            await tool_manager.get_tool(name)

    @pytest.mark.asyncio
    async def test_remove_nonexistent_tool(self, tool_manager):
        """Test removing non-existent tool."""
        with pytest.raises(NotFoundError):
            await tool_manager.remove_tool("NonExistent_tool")

    @pytest.mark.asyncio
    async def test_tool_conversion(self, tool_manager):
        """Test conversion of MaterializedTool to MCP Tool format."""
        tools = await tool_manager.list_tools()
        if not tools:
            pytest.skip("No tools in manager to validate conversion")
        tool = tools[0]

        # Check required fields
        assert isinstance(tool.name, str)
        assert isinstance(tool.description, str) or tool.description is None
        assert "inputSchema" in tool.model_dump()

        schema = tool.inputSchema
        assert schema["type"] == "object"
        assert "properties" in schema



"""
Tests for EvalSuite capture mode functionality.

Capture mode allows running evaluations without scoring - it simply records
the tool calls made by the model for debugging or generating expected calls.
"""

import json
import tempfile
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from arcade_evals import (
    CapturedCase,
    CapturedToolCall,
    CaptureResult,
    EvalSuite,
)

# Mark all tests in this module as requiring evals dependencies
pytestmark = pytest.mark.evals

# --- CapturedToolCall Tests ---


class TestCapturedToolCall:
    """Tests for CapturedToolCall dataclass."""

    def test_create_with_name_and_args(self):
        """Test creating a captured tool call with name and args."""
        tc = CapturedToolCall(name="Weather_GetCurrent", args={"location": "London"})
        assert tc.name == "Weather_GetCurrent"
        assert tc.args == {"location": "London"}

    def test_create_with_name_only(self):
        """Test creating a captured tool call with default empty args."""
        tc = CapturedToolCall(name="Weather_GetCurrent")
        assert tc.name == "Weather_GetCurrent"
        assert tc.args == {}

    def test_to_dict(self):
        """Test to_dict serialization."""
        tc = CapturedToolCall(name="MyTool", args={"key": "value"})
        result = tc.to_dict()
        assert result == {"name": "MyTool", "args": {"key": "value"}}

    def test_to_dict_empty_args(self):
        """Test to_dict with empty args."""
        tc = CapturedToolCall(name="MyTool")
        result = tc.to_dict()
        assert result == {"name": "MyTool", "args": {}}


# --- CapturedCase Tests ---


class TestCapturedCase:
    """Tests for CapturedCase dataclass."""

    def test_create_basic(self):
        """Test creating a captured case with minimal fields."""
        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[CapturedToolCall(name="Tool1")],
        )
        assert case.case_name == "test_case"
        assert case.user_message == "Hello"
        assert len(case.tool_calls) == 1
        assert case.system_message is None
        assert case.additional_messages is None

    def test_create_with_context(self):
        """Test creating a captured case with full context."""
        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[CapturedToolCall(name="Tool1")],
            system_message="You are an assistant",
            additional_messages=[{"role": "assistant", "content": "Hi"}],
        )
        assert case.system_message == "You are an assistant"
        assert case.additional_messages == [{"role": "assistant", "content": "Hi"}]

    def test_to_dict_without_context(self):
        """Test to_dict without including context."""
        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[CapturedToolCall(name="Tool1", args={"x": 1})],
            system_message="System message",
            additional_messages=[{"role": "user", "content": "msg"}],
        )
        result = case.to_dict(include_context=False)
        assert result == {
            "case_name": "test_case",
            "user_message": "Hello",
            "tool_calls": [{"name": "Tool1", "args": {"x": 1}}],
        }
        # Context should NOT be included
        assert "system_message" not in result
        assert "additional_messages" not in result

    def test_to_dict_with_context(self):
        """Test to_dict including context."""
        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[CapturedToolCall(name="Tool1", args={"x": 1})],
            system_message="System message",
            additional_messages=[{"role": "user", "content": "msg"}],
        )
        result = case.to_dict(include_context=True)
        assert result == {
            "case_name": "test_case",
            "user_message": "Hello",
            "tool_calls": [{"name": "Tool1", "args": {"x": 1}}],
            "system_message": "System message",
            "additional_messages": [{"role": "user", "content": "msg"}],
        }

    def test_to_dict_with_context_null_messages(self):
        """Test to_dict with context when additional_messages is None."""
        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[],
            system_message="Sys",
            additional_messages=None,
        )
        result = case.to_dict(include_context=True)
        assert result["additional_messages"] == []

    def test_to_dict_normalizes_json_string_arguments(self):
        """Test that JSON string arguments in additional_messages are parsed into objects."""
        # This simulates OpenAI's format where arguments is a JSON string
        additional_messages = [
            {"role": "user", "content": "List projects"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_123",
                        "type": "function",
                        "function": {
                            "name": "Linear_ListProjects",
                            "arguments": '{"state": "started"}',  # JSON string
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": '{"projects": []}',
                "tool_call_id": "call_123",
            },
        ]

        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[],
            system_message="Sys",
            additional_messages=additional_messages,
        )
        result = case.to_dict(include_context=True)

        # Arguments should be parsed into an object, not a string
        assistant_msg = result["additional_messages"][1]
        assert assistant_msg["tool_calls"][0]["function"]["arguments"] == {"state": "started"}

    def test_to_dict_handles_invalid_json_arguments(self):
        """Test that invalid JSON arguments are kept as strings."""
        additional_messages = [
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_123",
                        "type": "function",
                        "function": {
                            "name": "SomeTool",
                            "arguments": "not valid json {",  # Invalid JSON
                        },
                    }
                ],
            },
        ]

        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[],
            system_message="Sys",
            additional_messages=additional_messages,
        )
        result = case.to_dict(include_context=True)

        # Invalid JSON should remain as string
        assistant_msg = result["additional_messages"][0]
        assert assistant_msg["tool_calls"][0]["function"]["arguments"] == "not valid json {"

    def test_to_dict_normalizes_tool_response_content(self):
        """Test that JSON content in tool response messages is parsed into objects."""
        additional_messages = [
            {"role": "user", "content": "Get the initiative"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_get_init",
                        "type": "function",
                        "function": {
                            "name": "Linear_GetInitiative",
                            "arguments": '{"id": "init_123"}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": '{"id": "init_123", "name": "Q1 Goals", "status": "Planned"}',
                "tool_call_id": "call_get_init",
                "name": "Linear_GetInitiative",
            },
        ]

        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[],
            system_message="Sys",
            additional_messages=additional_messages,
        )
        result = case.to_dict(include_context=True)

        # Tool call arguments should be parsed
        assistant_msg = result["additional_messages"][1]
        assert assistant_msg["tool_calls"][0]["function"]["arguments"] == {"id": "init_123"}

        # Tool response content should be parsed
        tool_msg = result["additional_messages"][2]
        assert tool_msg["content"] == {"id": "init_123", "name": "Q1 Goals", "status": "Planned"}

    def test_to_dict_keeps_non_json_tool_content_as_string(self):
        """Test that non-JSON tool content is kept as string."""
        additional_messages = [
            {
                "role": "tool",
                "content": "Error: Tool not found",  # Plain text, not JSON
                "tool_call_id": "call_123",
                "name": "SomeTool",
            },
        ]

        case = CapturedCase(
            case_name="test_case",
            user_message="Hello",
            tool_calls=[],
            system_message="Sys",
            additional_messages=additional_messages,
        )
        result = case.to_dict(include_context=True)

        # Non-JSON content should remain as string
        tool_msg = result["additional_messages"][0]
        assert tool_msg["content"] == "Error: Tool not found"

    def test_empty_tool_calls(self):
        """Test case with no tool calls."""
        case = CapturedCase(
            case_name="no_tools",
            user_message="Just chat",
            tool_calls=[],
        )
        result = case.to_dict()
        assert result["tool_calls"] == []


# --- CaptureResult Tests ---


class TestCaptureResult:
    """Tests for CaptureResult dataclass."""

    def test_create_basic(self):
        """Test creating a capture result."""
        result = CaptureResult(
            suite_name="My Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[CapturedToolCall(name="Tool1")],
                )
            ],
        )
        assert result.suite_name == "My Suite"
        assert result.model == "gpt-4o"
        assert result.provider == "openai"
        assert len(result.captured_cases) == 1

    def test_to_dict_without_context(self):
        """Test to_dict without context."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[CapturedToolCall(name="Tool1", args={"a": 1})],
                    system_message="System",
                )
            ],
        )
        d = result.to_dict(include_context=False)
        assert d["suite_name"] == "Suite"
        assert d["model"] == "gpt-4o"
        assert d["provider"] == "openai"
        assert len(d["captured_cases"]) == 1
        assert "system_message" not in d["captured_cases"][0]

    def test_to_dict_with_context(self):
        """Test to_dict with context."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[],
                    system_message="System",
                    additional_messages=[],
                )
            ],
        )
        d = result.to_dict(include_context=True)
        assert d["captured_cases"][0]["system_message"] == "System"

    def test_to_json(self):
        """Test JSON serialization."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[CapturedToolCall(name="Tool1")],
                )
            ],
        )
        json_str = result.to_json(include_context=False)
        parsed = json.loads(json_str)
        assert parsed["suite_name"] == "Suite"
        assert parsed["model"] == "gpt-4o"

    def test_to_json_with_indent(self):
        """Test JSON serialization with custom indent."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[],
        )
        json_str = result.to_json(indent=4)
        # Check that indentation is present (4 spaces)
        assert "    " in json_str

    def test_write_to_file(self):
        """Test writing capture result to file."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[CapturedToolCall(name="Tool1", args={"x": 1})],
                )
            ],
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "capture_output.json"
            result.write_to_file(str(filepath))

            # Verify file was created and has valid content
            assert filepath.exists()
            with open(filepath, encoding="utf-8") as f:
                data = json.load(f)
            assert data["suite_name"] == "Suite"
            assert len(data["captured_cases"]) == 1

    def test_write_to_file_with_context(self):
        """Test writing capture result with context to file."""
        result = CaptureResult(
            suite_name="Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Hello",
                    tool_calls=[],
                    system_message="System",
                    additional_messages=[{"role": "user", "content": "x"}],
                )
            ],
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "capture_output.json"
            result.write_to_file(str(filepath), include_context=True)

            with open(filepath, encoding="utf-8") as f:
                data = json.load(f)
            assert data["captured_cases"][0]["system_message"] == "System"

    def test_empty_captured_cases(self):
        """Test with no captured cases."""
        result = CaptureResult(
            suite_name="Empty Suite",
            model="gpt-4o",
            provider="openai",
            captured_cases=[],
        )
        d = result.to_dict()
        assert d["captured_cases"] == []


# --- Imports Test ---


class TestCaptureImports:
    """Tests for capture mode imports."""

    def test_import_from_arcade_evals(self):
        """Test that capture classes are importable from arcade_evals."""
        from arcade_evals import CapturedCase, CapturedToolCall, CaptureResult

        assert CapturedToolCall is not None
        assert CapturedCase is not None
        assert CaptureResult is not None


# --- EvalSuite.capture() Tests ---


class TestEvalSuiteCapture:
    """Tests for EvalSuite.capture() method."""

    @pytest.fixture
    def mock_openai_client(self):
        """Create a mock OpenAI client."""
        client = AsyncMock()
        # Create mock response with tool calls
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]
        mock_response.choices[0].message.tool_calls = [MagicMock()]
        mock_response.choices[0].message.tool_calls[0].function.name = "Weather_GetCurrent"
        mock_response.choices[0].message.tool_calls[0].function.arguments = '{"location": "London"}'
        client.chat.completions.create = AsyncMock(return_value=mock_response)
        return client

    @pytest.fixture
    def basic_suite(self):
        """Create a basic EvalSuite with a tool and case."""
        suite = EvalSuite(
            name="Test Suite",
            system_message="You are a helpful assistant",
        )
        suite.add_tool_definitions([
            {"name": "Weather_GetCurrent", "description": "Get weather", "inputSchema": {}}
        ])
        suite.add_case(
            name="test_case",
            user_message="What's the weather in London?",
            expected_tool_calls=[],  # No expectations in capture mode
        )
        return suite

    @pytest.mark.asyncio
    async def test_capture_returns_capture_result(self, basic_suite, mock_openai_client):
        """Test that capture() returns a CaptureResult."""
        result = await basic_suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
        )
        assert isinstance(result, CaptureResult)
        assert result.suite_name == "Test Suite"
        assert result.model == "gpt-4o"
        assert result.provider == "openai"

    @pytest.mark.asyncio
    async def test_capture_records_tool_calls(self, basic_suite, mock_openai_client):
        """Test that capture() records tool calls from model."""
        result = await basic_suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
        )
        assert len(result.captured_cases) == 1
        case = result.captured_cases[0]
        assert case.case_name == "test_case"
        assert len(case.tool_calls) == 1
        assert case.tool_calls[0].name == "Weather_GetCurrent"
        assert case.tool_calls[0].args == {"location": "London"}

    @pytest.mark.asyncio
    async def test_capture_without_context(self, basic_suite, mock_openai_client):
        """Test that capture() without context doesn't include system message."""
        result = await basic_suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
            include_context=False,
        )
        case = result.captured_cases[0]
        assert case.system_message is None
        assert case.additional_messages is None

    @pytest.mark.asyncio
    async def test_capture_with_context(self, basic_suite, mock_openai_client):
        """Test that capture() with context includes system message."""
        result = await basic_suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
            include_context=True,
        )
        case = result.captured_cases[0]
        assert case.system_message == "You are a helpful assistant"
        assert case.additional_messages is not None

    @pytest.mark.asyncio
    async def test_capture_requires_tools(self):
        """Test that capture() raises error when no tools registered."""
        suite = EvalSuite(
            name="Empty Suite",
            system_message="Test",
        )
        suite.add_case(
            name="test_case",
            user_message="Hello",
            expected_tool_calls=[],
        )

        mock_client = AsyncMock()
        with pytest.raises(ValueError, match="No tools registered"):
            await suite.capture(mock_client, "gpt-4o", provider="openai")

    @pytest.mark.asyncio
    async def test_capture_multiple_cases(self, mock_openai_client):
        """Test capture with multiple cases."""
        suite = EvalSuite(
            name="Multi Case Suite",
            system_message="You are an assistant",
        )
        suite.add_tool_definitions([
            {"name": "Tool1", "description": "Tool 1"},
            {"name": "Tool2", "description": "Tool 2"},
        ])
        suite.add_case(name="case1", user_message="Do thing 1", expected_tool_calls=[])
        suite.add_case(name="case2", user_message="Do thing 2", expected_tool_calls=[])

        result = await suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
        )
        assert len(result.captured_cases) == 2
        assert result.captured_cases[0].case_name == "case1"
        assert result.captured_cases[1].case_name == "case2"

    @pytest.mark.asyncio
    async def test_capture_no_tool_calls(self):
        """Test capture when model doesn't call any tools."""
        suite = EvalSuite(
            name="No Calls Suite",
            system_message="Test",
        )
        suite.add_tool_definitions([{"name": "Tool1", "description": "Tool 1"}])
        suite.add_case(name="case1", user_message="Hello", expected_tool_calls=[])

        # Mock client that returns no tool calls
        mock_client = AsyncMock()
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]
        mock_response.choices[0].message.tool_calls = None
        mock_client.chat.completions.create = AsyncMock(return_value=mock_response)

        result = await suite.capture(
            client=mock_client,
            model="gpt-4o",
            provider="openai",
        )
        assert len(result.captured_cases) == 1
        assert len(result.captured_cases[0].tool_calls) == 0

    @pytest.mark.asyncio
    async def test_capture_normalizes_tool_calls(self, mock_openai_client):
        """Test that capture() normalizes tool names and fills defaults."""
        suite = EvalSuite(
            name="Normalization Suite",
            system_message="Test",
        )
        # Add tool with default arg
        suite.add_tool_definitions([
            {
                "name": "My.Tool",
                "description": "Tool with default",
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "arg1": {"type": "string", "default": "default_val"},
                        "arg2": {"type": "string"},
                    },
                },
            }
        ])
        suite.add_case(name="case1", user_message="Call it", expected_tool_calls=[])

        # Mock client returning tool call with underscored name and missing default arg
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]
        mock_response.choices[0].message.tool_calls = [MagicMock()]

        tool_call = mock_response.choices[0].message.tool_calls[0]
        tool_call.function.name = "My_Tool"  # Normalized name
        tool_call.function.arguments = '{"arg2": "provided"}'  # Missing arg1

        mock_openai_client.chat.completions.create = AsyncMock(return_value=mock_response)

        result = await suite.capture(
            client=mock_openai_client,
            model="gpt-4o",
            provider="openai",
        )

        case = result.captured_cases[0]
        # Name is resolved to original format (My_Tool -> My.Tool)
        # This ensures consistency with expected tool names
        assert case.tool_calls[0].name == "My.Tool"
        # Args should include default value
        assert case.tool_calls[0].args == {"arg1": "default_val", "arg2": "provided"}


# --- tool_eval decorator capture mode Tests ---


class TestToolEvalCaptureMode:
    """Tests for tool_eval decorator with capture mode."""

    @pytest.mark.asyncio
    async def test_tool_eval_capture_mode_flag(self):
        """Test that tool_eval wrapper passes capture_mode correctly."""
        from arcade_evals import tool_eval

        @tool_eval()
        def my_eval():
            suite = EvalSuite(
                name="Test Suite",
                system_message="Test",
            )
            suite.add_tool_definitions([{"name": "Tool1", "description": "D"}])
            suite.add_case(name="case1", user_message="Hello", expected_tool_calls=[])
            return suite

        # Mock the underlying capture functions
        with patch("arcade_evals.eval._capture_with_openai") as mock_capture:
            mock_capture.return_value = CaptureResult(
                suite_name="Test",
                model="gpt-4o",
                provider="openai",
                captured_cases=[],
            )

            results = await my_eval(
                provider_api_key="test-key",
                model="gpt-4o",
                capture_mode=True,
                include_context=False,
            )

            mock_capture.assert_called_once()
            assert len(results) == 1
            assert isinstance(results[0], CaptureResult)

    @pytest.mark.asyncio
    async def test_tool_eval_capture_mode_with_context(self):
        """Test that tool_eval wrapper passes include_context correctly."""
        from arcade_evals import tool_eval

        @tool_eval()
        def my_eval():
            suite = EvalSuite(
                name="Test Suite",
                system_message="Test",
            )
            suite.add_tool_definitions([{"name": "Tool1", "description": "D"}])
            suite.add_case(name="case1", user_message="Hello", expected_tool_calls=[])
            return suite

        with patch("arcade_evals.eval._capture_with_openai") as mock_capture:
            mock_capture.return_value = CaptureResult(
                suite_name="Test",
                model="gpt-4o",
                provider="openai",
                captured_cases=[],
            )

            await my_eval(
                provider_api_key="test-key",
                model="gpt-4o",
                capture_mode=True,
                include_context=True,
            )

            # Verify include_context was passed
            call_args = mock_capture.call_args
            assert call_args.kwargs["include_context"] is True


# --- Multiple Tool Calls per Case Tests ---


class TestMultipleToolCalls:
    """Tests for capturing multiple tool calls from a single case."""

    @pytest.mark.asyncio
    async def test_capture_multiple_tool_calls(self):
        """Test capturing multiple tool calls from one model response."""
        suite = EvalSuite(
            name="Multi Tool Suite",
            system_message="Test",
        )
        suite.add_tool_definitions([
            {"name": "Tool1", "description": "D1"},
            {"name": "Tool2", "description": "D2"},
        ])
        suite.add_case(name="case1", user_message="Do both", expected_tool_calls=[])

        # Mock client returning multiple tool calls
        mock_client = AsyncMock()
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]

        tool_call_1 = MagicMock()
        tool_call_1.function.name = "Tool1"
        tool_call_1.function.arguments = '{"arg1": "val1"}'

        tool_call_2 = MagicMock()
        tool_call_2.function.name = "Tool2"
        tool_call_2.function.arguments = '{"arg2": "val2"}'

        mock_response.choices[0].message.tool_calls = [tool_call_1, tool_call_2]
        mock_client.chat.completions.create = AsyncMock(return_value=mock_response)

        result = await suite.capture(
            client=mock_client,
            model="gpt-4o",
            provider="openai",
        )

        assert len(result.captured_cases) == 1
        case = result.captured_cases[0]
        assert len(case.tool_calls) == 2
        assert case.tool_calls[0].name == "Tool1"
        assert case.tool_calls[0].args == {"arg1": "val1"}
        assert case.tool_calls[1].name == "Tool2"
        assert case.tool_calls[1].args == {"arg2": "val2"}


class TestCaptureWithAnthropic:
    """Tests for capture mode with Anthropic provider."""

    @pytest.mark.asyncio
    async def test_capture_with_anthropic_provider(self):
        """Test capture mode using Anthropic provider."""
        suite = EvalSuite(
            name="Anthropic Capture Suite",
            system_message="Test system message",
        )
        suite.add_tool_definitions([
            {"name": "Google.Search", "description": "Search"},
        ])
        suite.add_case(
            name="test_case",
            user_message="Search for something",
            expected_tool_calls=[],
        )

        # Mock Anthropic client
        mock_client = AsyncMock()
        mock_response = MagicMock()

        # Anthropic returns tool_use blocks
        mock_tool_use = MagicMock()
        mock_tool_use.type = "tool_use"
        mock_tool_use.name = "Google_Search"  # Anthropic uses underscores
        mock_tool_use.input = {"query": "test"}

        mock_response.content = [mock_tool_use]
        mock_client.messages.create = AsyncMock(return_value=mock_response)

        result = await suite.capture(
            client=mock_client,
            model="claude-3-opus",
            provider="anthropic",
        )

        assert result.provider == "anthropic"
        assert len(result.captured_cases) == 1
        # Should resolve Google_Search back to Google.Search
        assert result.captured_cases[0].tool_calls[0].name == "Google.Search"


class TestCaptureHelperFunctions:
    """Tests for _capture_with_openai and _capture_with_anthropic helpers."""

    @pytest.mark.asyncio
    async def test_capture_with_openai_helper(self):
        """Test the _capture_with_openai helper function."""
        from arcade_evals.capture import _capture_with_openai

        suite = EvalSuite(
            name="OpenAI Helper Test",
            system_message="Test",
        )
        suite.add_tool_definitions([{"name": "TestTool", "description": "A test tool"}])
        suite.add_case(name="case1", user_message="Test", expected_tool_calls=[])

        # Mock the suite.capture method directly instead of AsyncOpenAI
        mock_result = CaptureResult(
            suite_name="OpenAI Helper Test",
            provider="openai",
            model="gpt-4o",
            captured_cases=[
                CapturedCase(
                    case_name="case1",
                    user_message="Test",
                    tool_calls=[],
                    system_message="Test",
                    additional_messages=[],
                )
            ],
        )

        with patch.object(suite, "capture", return_value=mock_result) as mock_capture:
            result = await _capture_with_openai(
                suite=suite,
                api_key="test-key",
                model="gpt-4o",
                include_context=True,
            )

            assert result.suite_name == "OpenAI Helper Test"
            assert result.provider == "openai"
            # Verify capture was called with correct arguments
            mock_capture.assert_called_once()
            call_args = mock_capture.call_args
            # Arguments: (client, model, provider=..., include_context=...)
            assert call_args.args[1] == "gpt-4o"  # model
            assert call_args.kwargs.get("provider") == "openai"
            assert call_args.kwargs.get("include_context") is True

    def test_capture_with_anthropic_function_exists(self):
        """Test that _capture_with_anthropic helper function exists and is callable."""
        # Verify the function exists and has the expected signature
        import inspect

        from arcade_evals.capture import _capture_with_anthropic

        sig = inspect.signature(_capture_with_anthropic)
        params = list(sig.parameters.keys())
        assert "suite" in params
        assert "api_key" in params
        assert "model" in params
        assert "include_context" in params



from pathlib import Path


def get_minimal_template_directory() -> Path:
    """Get the path to the templates directory."""
    return Path(__file__).parent / "minimal"

def get_full_template_directory() -> Path:
    """Get the path to the templates directory."""
    return Path(__file__).parent / "full"



"""Tests for track management in comparative evaluations."""

import pytest
from arcade_evals._evalsuite._tool_registry import EvalSuiteToolRegistry
from arcade_evals._evalsuite._tracks import TrackManager

# Mark all tests in this module as requiring evals dependencies
pytestmark = pytest.mark.evals


class TestTrackManager:
    """Tests for TrackManager class."""

    def test_create_track(self) -> None:
        """Test creating a new track."""
        manager = TrackManager()
        registry = EvalSuiteToolRegistry()

        track_name = manager.create_track("Test Track", registry)

        assert track_name == "Test Track"
        assert manager.has_track("Test Track")
        assert manager.track_count() == 1

    def test_create_duplicate_track_raises(self) -> None:
        """Test that creating a duplicate track raises ValueError."""
        manager = TrackManager()
        registry1 = EvalSuiteToolRegistry()
        registry2 = EvalSuiteToolRegistry()

        manager.create_track("Track1", registry1)

        with pytest.raises(ValueError, match="already exists"):
            manager.create_track("Track1", registry2)

    def test_get_registry(self) -> None:
        """Test getting a registry by track name."""
        manager = TrackManager()
        registry = EvalSuiteToolRegistry()
        registry.add_tool({"name": "TestTool", "description": "Test"})

        manager.create_track("MyTrack", registry)
        retrieved = manager.get_registry("MyTrack")

        assert retrieved is registry
        assert retrieved.has_tool("TestTool")

    def test_get_registry_nonexistent(self) -> None:
        """Test getting a nonexistent registry returns None."""
        manager = TrackManager()

        result = manager.get_registry("NonexistentTrack")

        assert result is None

    def test_get_track_names(self) -> None:
        """Test getting all track names."""
        manager = TrackManager()
        manager.create_track("Track1", EvalSuiteToolRegistry())
        manager.create_track("Track2", EvalSuiteToolRegistry())
        manager.create_track("Track3", EvalSuiteToolRegistry())

        names = manager.get_track_names()

        assert names == ["Track1", "Track2", "Track3"]

    def test_get_track_names_empty(self) -> None:
        """Test getting track names when empty."""
        manager = TrackManager()

        names = manager.get_track_names()

        assert names == []

    def test_has_track(self) -> None:
        """Test checking if track exists."""
        manager = TrackManager()
        manager.create_track("Exists", EvalSuiteToolRegistry())

        assert manager.has_track("Exists") is True
        assert manager.has_track("DoesNotExist") is False

    def test_track_count(self) -> None:
        """Test counting tracks."""
        manager = TrackManager()

        assert manager.track_count() == 0

        manager.create_track("Track1", EvalSuiteToolRegistry())
        assert manager.track_count() == 1

        manager.create_track("Track2", EvalSuiteToolRegistry())
        assert manager.track_count() == 2

    def test_get_all_registries(self) -> None:
        """Test getting all registries."""
        manager = TrackManager()
        reg1 = EvalSuiteToolRegistry()
        reg2 = EvalSuiteToolRegistry()

        manager.create_track("Track1", reg1)
        manager.create_track("Track2", reg2)

        all_regs = manager.get_all_registries()

        assert len(all_regs) == 2
        assert all_regs["Track1"] is reg1
        assert all_regs["Track2"] is reg2

    def test_registries_are_isolated(self) -> None:
        """Test that each track has its own isolated registry."""
        manager = TrackManager()
        reg1 = EvalSuiteToolRegistry()
        reg2 = EvalSuiteToolRegistry()

        reg1.add_tool({"name": "Tool1", "description": "Tool for track 1"})
        reg2.add_tool({"name": "Tool2", "description": "Tool for track 2"})

        manager.create_track("Track1", reg1)
        manager.create_track("Track2", reg2)

        # Each registry only has its own tool
        assert manager.get_registry("Track1").has_tool("Tool1")
        assert not manager.get_registry("Track1").has_tool("Tool2")
        assert manager.get_registry("Track2").has_tool("Tool2")
        assert not manager.get_registry("Track2").has_tool("Tool1")
